/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.communication
 * File:    Client.java
 *
 * Copyright (C) 2010 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.communication;

import java.io.IOException;
import java.io.InputStream;
import java.net.Socket;
import java.security.NoSuchAlgorithmException;

import javax.xml.bind.JAXBException;

import org.apache.http.HeaderElement;
import org.apache.http.HeaderElementIterator;
import org.apache.http.HttpEntity;
import org.apache.http.HttpEntityEnclosingRequest;
import org.apache.http.HttpException;
import org.apache.http.HttpRequest;
import org.apache.http.HttpResponse;
import org.apache.http.HttpStatus;
import org.apache.http.HttpVersion;
import org.apache.http.entity.InputStreamEntity;
import org.apache.http.entity.StringEntity;
import org.apache.http.impl.DefaultHttpServerConnection;
import org.apache.http.message.BasicHeader;
import org.apache.http.message.BasicHeaderElementIterator;
import org.apache.http.message.BasicHttpResponse;
import org.apache.http.params.BasicHttpParams;
import org.apache.log4j.Logger;
import org.trustedcomputinggroup._2010.ifmap._2.ResponseType;
import org.w3._2003._05.soap_envelope.Body;
import org.w3._2003._05.soap_envelope.Envelope;
import org.w3._2003._05.soap_envelope.Header;

import de.fhhannover.inform.iron.mapserver.communication.connections.ConnectionContainer;
import de.fhhannover.inform.iron.mapserver.communication.connections.SSRC;
import de.fhhannover.inform.iron.mapserver.communication.services.ServiceHandler;
import de.fhhannover.inform.iron.mapserver.communication.services.ssrc.exceptions.SessionNotFoundException;
import de.fhhannover.inform.iron.mapserver.logging.Logging;

public abstract class Client implements Runnable {
	private Socket mSocket;
	private MessageFactory mMf;
	private DefaultHttpServerConnection mServerConn;
	private Boolean ssrc;
	private String publisher = null;
	protected static Logger logger;
	static {
		logger = Logging.getTheLogger();
	}
	
	public Client(Socket socket) throws JAXBException {
		mSocket = socket;
		mMf = new MessageFactory();
	}
	
	public void setSocket(Socket socket){
		mSocket = socket;
	}
	
	@Override
	public void run() {
		
		try {
			initDefaultHttpServerConnection();
		
			try {
				doMessageExchange();
			} catch (NoSuchAlgorithmException e) {
				//Nothing to do
				//Socket is already closed. No response possible
				logger.debug("No such Algorithm. Server Error");
			} catch (SessionNotFoundException e) {
				logger.debug("SessionNotFound. Server Error");
				//Nothing to do
				//Socket is already closed. No response possible	
			} catch (IllegalStateException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (JAXBException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			
		} catch (IOException e) {
			logger.debug("IOException occured");
			e.printStackTrace();
		} catch (HttpException e) {
			logger.debug("HTTPException occured");
			e.printStackTrace();
		}
	}
	// getHeader
	// check autorized
	// processContent
	private void doMessageExchange() throws HttpException, NoSuchAlgorithmException, SessionNotFoundException, IllegalStateException, JAXBException {
		
		while(true) {
			logger.trace("receiving request header");
			
			HttpRequest req;
			try {
				req = mServerConn.receiveRequestHeader();
	
				
				HttpEntity reqEntity = null;
				if (req instanceof HttpEntityEnclosingRequest) {
					mServerConn.receiveRequestEntity((HttpEntityEnclosingRequest) req);
					reqEntity = ((HttpEntityEnclosingRequest)req).getEntity();
					logger.debug("Received Request Entity");
				}
	
				HeaderElementIterator it = new BasicHeaderElementIterator(
						req.headerIterator("Authorization"));
				
				boolean authTemp = isAuthorized();
				
				while (it.hasNext() && !authTemp) {
					logger.info("in while(): checking authorization");
					HeaderElement el = it.nextElement();
					authTemp = authorize(el.getName());
				}
				
				if (!authTemp) {
					logger.debug("Client not authorized! Sending 401");
					sendUnauthorizedResponse();
					if (reqEntity != null) {
						reqEntity.consumeContent();
					}
					continue;
				}

				publisher = determinePublisherId();
				
				if (reqEntity != null) {
					
					Envelope env = mMf.getRequest(reqEntity.getContent());
					Envelope resp = null;;
					
					if (env.getBody().getPoll() != null) {
							if (ssrc != null && ssrc == true) {
								logger.warn("POLL REQUEST ON A SSRC DETECTED... CLOSING SOCKET!!!");
								// TODO: Do whatever the spec says should be done,
								// for now we simply close 'em away...
								// Interestingly that does the trick for soap ui
								// and axis as it seems they will reconnect...
								// However it probably screwed the timer stuff
								mSocket.close();
								throw new IOException("poll on ssrc");	// start a timer?
							} else {
								ssrc = false;
								logger.trace("Serving poll request on ARC channel");
								processARCRequest(env, publisher);
							}
					} else {
						if (ssrc == null) {
							logger.trace("New SSRC channel for publisher=" + publisher);
							ssrc = true;
						}
						if (ssrc == true) {
							logger.trace("Serving request on SSRC channel");
							resp = processSSRCRequestToResponse(env, publisher);
							sendResponse(resp);
						} else {
							logger.warn("SSRC REQUEST ON AN ARC DETECTED... CLOSING SOCKET!");
							// TODO: Do whatever the spec says should be done,
							// for now we simply close 'em away...
							// Interestingly that does the trick for soap ui
							// and axis as it seems they will reconnect...
							// However it probably screwed the timer stuff
							mSocket.close();
							ssrc = false;
							throw new IOException("ssrc on arc");	// start a timer?
						}
					}
					logger.trace("Finished processing client request");
					
				} else {
					logger.info("Did not get any content of client...");
					sendNoContent();
				}
			} catch (IOException e) {
				if (e.getMessage().equals("Client closed connection")) {
						logger.info("Client from " + mSocket.getInetAddress().getHostAddress() +
								":" + mSocket.getPort() + " closed connection");
					if (publisher != null && ssrc != null && ssrc) {
						logger.info(" Starting Timer for publisher: "+ publisher);
						SSRC ssrc = ConnectionContainer.getInstance().getSessionByPublisherID(publisher);
						ssrc.startTimer(180);
					}
				} else {
					logger.warn("IO Exception");
					//e.printStackTrace();
				}
				break;
			}
		}
	}
	
	public void sendResponse(Envelope resp) throws HttpException, IOException, JAXBException {
		logger.trace("sending response out");
		
		InputStream respStream = mMf.getResponse(resp);
		int size = mMf.getResponseSize();

		BasicHeader header = new BasicHeader("Content-Length", "" + size);
		HttpResponse response = new BasicHttpResponse(HttpVersion.HTTP_1_1, HttpStatus.SC_OK, "");
		response.addHeader(header);
		HttpEntity respEntity = new InputStreamEntity(respStream, size);
		response.setEntity(respEntity);
		mServerConn.sendResponseHeader(response);
		mServerConn.sendResponseEntity(response);
		respEntity.consumeContent();
	}

	private void sendUnauthorizedResponse() throws HttpException, IOException {
		
		String html = "<html><head><title>401 Unauthorized</title></head></html>";
		HttpResponse denied = new BasicHttpResponse(HttpVersion.HTTP_1_1,
				HttpStatus.SC_UNAUTHORIZED, "Unauthorized");
		BasicHeader header = new BasicHeader("WWW-Authenticate", "Basic realm=\"VARIABLE MACHEN\"");

		denied.addHeader(header);
		header = new BasicHeader("Content-length", "" + html.length());
		denied.addHeader(header);
		
		StringEntity entity = new StringEntity(html);
		denied.setEntity(entity);
		
		mServerConn.sendResponseHeader(denied);
		mServerConn.sendResponseEntity(denied);
	}
	
	//TODO
	private void sendNoContent() throws HttpException, IOException {
		String html = "<html><head><title>402 Payment!!!</title></head><body>PAYMENT REQUIRED!</body></html>";
		
		HttpResponse denied = new BasicHttpResponse(HttpVersion.HTTP_1_1,
				HttpStatus.SC_PAYMENT_REQUIRED, "Please");
		
		BasicHeader header = new BasicHeader("Content-length", "" + html.length());
		denied.addHeader(header);

		
		StringEntity entity = new StringEntity(html);
		denied.setEntity(entity);
		mServerConn.sendResponseHeader(denied);
		mServerConn.sendResponseEntity(denied);
	}

	private void initDefaultHttpServerConnection() throws IOException {
		BasicHttpParams params = new BasicHttpParams();
		mServerConn = new DefaultHttpServerConnection();
		mServerConn.bind(mSocket, params);
	}
					
	abstract protected boolean authorize(String value);
	
	abstract protected String determinePublisherId();

	abstract protected boolean isAuthorized();

	private Envelope processSSRCRequestToResponse(Envelope env, String publisher) {
		 Envelope response = new Envelope();
		 Body bodyResp = new Body();
		 response.setBody(bodyResp);
		 response.setHeader(new Header());
		 
		 Body bodyReq = env.getBody();
		 
		 ResponseType resp = null;
		 logger.trace("Dispatching SSRC operation for publisher " + publisher);
		 
		 if (bodyReq.getNewSession() != null) {
			 logger.trace("NewSession");
			 resp = ServiceHandler.newSession(bodyReq.getNewSession(), publisher);
		 } else if (bodyReq.getPublish() != null) {
			 logger.trace("Publish");
			 resp = ServiceHandler.publish((bodyReq.getPublish()));
		 } else if (bodyReq.getEndSession() != null) {
			 logger.trace("EndSession");
			 resp = ServiceHandler.endSession(bodyReq.getEndSession());
		 } else if (bodyReq.getPurgePublisher() != null) {
			 logger.trace("PurgePublisher");
			 resp = ServiceHandler.purgePublisher(bodyReq.getPurgePublisher());
		 } else if (bodyReq.getSearch() != null) {
			 logger.trace("Search");
			 resp = ServiceHandler.search(bodyReq.getSearch());
		 } else if (bodyReq.getSubscribe() != null) {
			 resp = ServiceHandler.subscribe(bodyReq.getSubscribe());
			 logger.trace("Subscribe");
		 } else if (bodyReq.getRenewSession() != null) {
			 logger.trace("RenewSession");
			 resp = ServiceHandler.renewSession(bodyReq.getRenewSession());
		 } else if (bodyReq.getAttachSession() != null) {
			 logger.fatal("INVALID AttachSession");
			 //resp = ServiceHandler.attachSession(bodyReq.getAttachSession());
			 
		 }else if (bodyReq.getDump() != null) {	
			logger.info("DumpRequest erhalten.");
			resp = ServiceHandler.dump(bodyReq.getDump());
		 }
		
		if (resp != null) {
			bodyResp.setResponse(resp);			
		} else {
			logger.error("bodyResp is null!");
		}
		
		return response;
	}
	
	private void processARCRequest(Envelope env, String publisher) {
		logger.trace("processing poll request");
		 Envelope response = new Envelope();
		 Body bodyResp = new Body();
		 response.setBody(bodyResp);
		 response.setHeader(new Header());
		 
		 Body bodyReq = env.getBody();
		 if (bodyReq.getPoll() != null) {
			ServiceHandler.poll(bodyReq.getPoll(), this);
		 } else {
			 logger.warn("In processARCRequest, but getPoll() is null");
		 }
	}

	public Socket getSocket(){
		return mSocket;
	}
}