/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.configuration
 * File:    ConfigurationReader.java
 *
 * Copyright (C) 2010 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.configuration;

import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.util.Properties;

import org.apache.log4j.Logger;

import de.fhhannover.inform.iron.mapserver.logging.Logging;

/**
 * The configuration reader parses the <code>ifmap.properties</code> file. The
 * configuration entries are stored as appropriate properties. irond supports
 * the following configuration options:</br>
 * 
 * <ul>
 * <li>identifier.casesensitive.hipHint=true</li>
 * <li>identifier.casesensitive.kerberosPrincipal=true</li>
 * <li>identifier.casesensitive.telUri=true</li>
 * <li>identifier.casesensitive.distinguishedName=true</li>
 * <li>defaultpollsize=5000</li>
 * <li>request.purge.restrict=true</li>
 * <li>identifier.casesensitive.aikName=true</li>
 * <li>identifier.casesensitive.userName=true</li>
 * <li>request.delete.restrict=true</li>
 * <li>identifier.casesensitive.emailAddress=true</li>
 * <li>identifier.casesensitive.sipUri=true</li>
 * <li>identifier.casesensitive.trustedPlatformModule=true</li>
 * <li>identifier.casesensitive.other=true</li>
 * <li>identifier.casesensitive.dnsName=true</li>
 * <li>identifier.casesensitive.administrativedomain=true</li>
 * <li>truststorePath=./keystore/mapsks.jks</li>
 * <li>keystorePath=./keystore/mapsks.jks</li>
 * <li>keystorePW=mapserver</li>
 * <li>truststorePW=mapserver</li>
 * <li>BasicAuthPort=8443</li>
 * <li>CredentialAuthPort=8444</li>
 * </ul>
 * 
 * @author Waldemar Bender
 * @author Dennis Ludewig
 * @author ibente
 * 
 */
public class ConfigurationReader extends Properties {

	private static final long serialVersionUID = 55304285127409015L;

	/**
	 * Singleton
	 */
	private static ConfigurationReader mConfReader;

	/**
	 * We expect the properties file to reside in the current working dir
	 */
	private static final String PROP_PATH = "ifmap.properties";

	/**
	 * The logger
	 */
	private static Logger logger;

	static {
		logger = Logging.getTheLogger();
	}

	/**
	 * Default constructor. The irond process will stop if the
	 * <code>ifmap.properties</code> configuration file can no be parsed.
	 */
	private ConfigurationReader() {
		try {
			logger.debug("current working dir is: "
					+ System.getProperty("user.dir"));
			FileReader mPropFile = new FileReader(PROP_PATH);
			this.load(mPropFile);
			mPropFile.close();

		} catch (FileNotFoundException e) {
			logger.warn("could not find ifmap.properties file in current directory, exiting now.");
			e.printStackTrace();
			System.exit(0);
		} catch (IOException e) {
			logger.warn("IO error when reading ifmap.properties file, exiting now.");
			e.printStackTrace();
			System.exit(0);
		}
	}

	/**
	 * Singleton
	 * 
	 * @return
	 */
	public static synchronized ConfigurationReader getInstance() {
		if (mConfReader == null) {
			mConfReader = new ConfigurationReader();
		}
		return mConfReader;
	}

	/**
	 * Adds a new key value pair and a comment to the current configuration and
	 * stores it in the <code>ifmap.properties</code> file.
	 * 
	 * @param key
	 * @param value
	 * @param comment
	 */
	public void put(String key, String value, String comment) {
		super.put(key, value);
		try {
			FileOutputStream fos = new FileOutputStream(PROP_PATH);
			this.store(fos, comment);
			fos.close();
		} catch (IOException e) {
			logger.warn("IO error during write to ifmap.properties, ignoring update: "
					+ key + "=" + value + " #" + comment);
			e.printStackTrace();
		}
	}

	/**
	 * Adds a new key value pair to the current configuration and stores it in
	 * the <code>ifmap.properties</code> file.
	 * 
	 * @param key
	 * @param value
	 */
	public void put(String key, String value) {
		super.put(key, value);
		try {
			FileOutputStream fos = new FileOutputStream(PROP_PATH);
			this.store(fos, "no comment");
			fos.close();
		} catch (IOException e) {
			logger.warn("IO error during write to ifmap.properties, ignoring update: "
					+ key + "=" + value);
			e.printStackTrace();
		}
	}
}
